@@ -1,5 +1,4 @@
-"""
-Extension functions for parsing sys.argv.
+"""Extension functions for parsing sys.argv.
 
 Commands:
 
@@ -9,24 +8,26 @@ Commands:
 
 from twill import log, namespaces, shell
 
-__all__ = ['get_args']
+__all__ = ["get_args"]
 
 
-def get_args(require=0):
+def get_args(require: int = 0) -> None:
     """>> get_args [<require>]
 
     Load the command line arguments after the last '--' into $arg1...$argN,
     optionally requiring at least 'require' such arguments.
     """
-    global_dict, local_dict = namespaces.get_twill_glocals()
+    global_dict, _local_dict = namespaces.get_twill_glocals()
 
     require = int(require)
 
     if len(shell.twill_args) < require:
         from twill.errors import TwillAssertionError
+
         given = len(shell.twill_args)
         raise TwillAssertionError(
-            f"too few arguments; {given} rather than {require}")
+            f"too few arguments; {given} rather than {require}"
+        )
 
     if shell.twill_args:
         for n, arg in enumerate(shell.twill_args, 1):
